@@ -8,6 +8,7 @@ import torchaudio.functional as AF
 import math
 import logging
 
+device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
 
 # Reference: https://github.com/jameslyons/python_speech_features
 # The style of this module mimics the `python_speech_features`
@@ -16,7 +17,7 @@ import logging
 # A details can be found in the reference.
 
 # Note that the precision of torch is float32 default. So there is a bit of gap between torch and numpy.
-
+# Focus on: input of a signal should be loaded on gpu.
 
 def preemphasis(signal, preemph=0.97):
     """
@@ -28,7 +29,7 @@ def preemphasis(signal, preemph=0.97):
     return torch.cat((signal[0:1], signal[1:] - preemph * signal[:-1]))
 
 
-def framesig(signal, framelen, framehop, winfunc=lambda x: torch.ones((x,))):
+def framesig(signal, framelen, framehop, winfunc=None):
     """
     Frame a signal into overlapping frames.
     :param signal: (time,)
@@ -38,48 +39,49 @@ def framesig(signal, framelen, framehop, winfunc=lambda x: torch.ones((x,))):
     :return: (nframes, framelen)
     """
     slen = len(signal)
-    framelen = round(framelen)  # round_half_up(framelen)
-    framehop = round(framehop)  # round_half_up(framehop)
+    framelen = round(framelen)
+    framehop = round(framehop)
     if slen <= framelen:
         nframes = 1
     else:
-        nframes = 1 + int(math.ceil((
-                                                1.0 * slen - framelen) / framehop))  # 1 + int(torch.ceil(torch.tensor((1.0 * slen - framelen) / framehop)))
+        nframes = 1 + int(math.ceil((1.0 * slen - framelen) / framehop))
 
     padlen = int((nframes - 1) * framehop + framelen)
 
-    zeros = torch.zeros((padlen - slen,))
+    zeros = torch.zeros((padlen - slen,), device=device)
     padsignal = torch.cat((signal, zeros))
 
-    indices = torch.arange(0, framelen).view((1, -1)) \
-              + torch.arange(0, nframes * framehop, framehop).view((-1, 1))
+    indices = torch.arange(0, framelen, device=device).view((1, -1)) \
+              + torch.arange(0, nframes * framehop, framehop, device=device).view((-1, 1))
     frames = padsignal[indices]
-    win = winfunc(framelen).view((1, -1))
-    return frames * win
+    if winfunc:
+        win = winfunc(framelen).view((1, -1))
+        frames = frames * win
+    return frames
 
 
 def magspec(frames, nfft):
     """
     Compute the magnitude spectrum of each frame.
+    The result of `torch.rfft` is separate pairs of real and complex.
     :param frames: (nframes, framelen)
     :param nfft:
     :return: (nframes, nfft//2+1)
     """
-    # torch.rfft 取窗口大小为 nfft，计算结果的实部与虚部分开
     if frames.shape[1] > nfft:
         logging.warning(
             'frame length (%d) is greater than FFT size (%d), frame will be truncated. Increase NFFT to avoid.',
             frames.shape[1], nfft)
         frames = frames[:, :nfft]
     else:
-        frames = torch.cat((frames, torch.zeros((frames.shape[0], nfft - frames.shape[1]))), 1)
+        frames = torch.cat((frames, torch.zeros((frames.shape[0], nfft - frames.shape[1]), device=device)), 1)
     complex_spec = torch.rfft(frames, 1)
     return torch.norm(complex_spec, dim=2)
 
 
 def powspec(frames, nfft):
     """
-    Compute the power spectrum of each fram.
+    Compute the power spectrum of each frame.
     :param frames: (nframes, framelen)
     :param nfft:
     :return: (nframes, nfft//2+1)
@@ -100,11 +102,10 @@ def calculate_nfft(samplerate, winlen):
         nfft <<= 1
     return nfft
 
-
 def mfcc(signal, samplerate=16000, winlen=0.025, hoplen=0.01,
          numcep=13, nfilt=26, nfft=None, lowfreq=0, highfreq=None,
-         preemph=0.97, ceplifter=22, plusEnergy=True,
-         winfunc=lambda x: torch.ones((x,))):
+         preemph=0.97, ceplifter=22, plusEnergy=True, dct=None,
+         winfunc=lambda x: torch.ones((x,), device=device)):
     """
     Compute MFCC from an audio signal.
     :param signal: (time,)
@@ -126,15 +127,17 @@ def mfcc(signal, samplerate=16000, winlen=0.025, hoplen=0.01,
     feat, energy = fbank(signal, samplerate, winlen, hoplen, nfilt,
                          nfft, lowfreq, highfreq, preemph, winfunc)
     feat = torch.log(feat)
-    feat = feat.mm(AF.create_dct(numcep, nfilt, norm='ortho'))
+    if not dct:
+        dct = AF.create_dct(numcep, nfilt, norm='ortho').to(device)
+    feat = feat.mm(dct)
     feat = lifter(feat, ceplifter)
     if plusEnergy: feat[:, 0] = torch.log(energy)
     return feat
 
 
 def fbank(signal, samplerate=16000, winlen=0.025, hoplen=0.01,
-          nfilt=26, nfft=512, lowfreq=0, highfreq=None,
-          preemph=0.97, winfunc=lambda x: torch.ones((x,))):
+          nfilt=26, nfft=None, lowfreq=0, highfreq=None,
+          preemph=0.97, winfunc=lambda x: torch.ones((x,), device=device)):
     """
     Compute Mel-filterbank energy features from an audio signal.
     :param signal: (time,)
@@ -149,6 +152,7 @@ def fbank(signal, samplerate=16000, winlen=0.025, hoplen=0.01,
     :param winfunc:
     :return: (nframes, nfilt), (frames,)
     """
+    nfft = nfft or calculate_nfft(samplerate, winlen)
     highfreq = highfreq or samplerate >> 1
     signal = preemphasis(signal, preemph)
     frames = framesig(signal, winlen * samplerate, hoplen * samplerate, winfunc)
@@ -160,12 +164,12 @@ def fbank(signal, samplerate=16000, winlen=0.025, hoplen=0.01,
     feat = torch.mm(pspec, fb.T)
     feat[feat == 0] = 2.220446049250313e-16  # 极小正数替换0
 
-    return feat, energy  # feat: size (winsize, nfilt)
+    return feat, energy  # feat: size (nframes, nfilt)
 
 
 def logfbank(signal, samplerate=16000, winlen=0.025, hoplen=0.01,
-             nfilt=26, nfft=512, lowfreq=0, highfreq=None,
-             preemph=0.97, winfunc=lambda x: torch.ones((x,))):
+             nfilt=26, nfft=None, lowfreq=0, highfreq=None,
+             preemph=0.97, winfunc=lambda x: torch.ones((x,), device=device)):
     """
     Compute log Mel-filterbank energy features from an audio signal as log(fbank).
     :param signal: (time,)
@@ -180,6 +184,7 @@ def logfbank(signal, samplerate=16000, winlen=0.025, hoplen=0.01,
     :param winfunc:
     :return: (nframes, nfilt)
     """
+    nfft = nfft or calculate_nfft(samplerate, winlen)
     feat, _ = fbank(signal, samplerate, winlen, hoplen, nfilt,
                     nfft, lowfreq, highfreq, preemph, winfunc)
     return torch.log(feat)
@@ -218,10 +223,10 @@ def get_filterbanks(nfilt=26, nfft=512, samplerate=16000, lowfreq=0, highfreq=No
 
     lowmel = hz2mel(lowfreq)
     highmel = hz2mel(highfreq)
-    melpoints = torch.linspace(lowmel, highmel, nfilt + 2)
+    melpoints = torch.linspace(lowmel, highmel, nfilt + 2, device=device)
     bin = torch.floor((nfft + 1) * mel2hz(melpoints) / samplerate)
 
-    fbank = torch.zeros((nfilt, (nfft >> 1) + 1))
+    fbank = torch.zeros((nfilt, (nfft >> 1) + 1), device=device)
     for j in range(0, nfilt):
         for i in range(int(bin[j]), int(bin[j + 1])):
             fbank[j, i] = (i - bin[j]) / (bin[j + 1] - bin[j])
@@ -232,14 +237,14 @@ def get_filterbanks(nfilt=26, nfft=512, samplerate=16000, lowfreq=0, highfreq=No
 
 def lifter(cepstra, ceplifter=22):
     """
-    Apply a cepstral lifer to the matrix of cepstra.
+    Apply a cepstral lifter to the matrix of cepstra.
     :param cepstra: (nframes, numcep)
     :param ceplifter:
     :return: (nframes, numcep)
     """
     if ceplifter > 0:
         nframes, numcep = cepstra.shape
-        lift = 1 + (ceplifter / 2.) * torch.sin(math.pi * torch.arange(numcep) / ceplifter)
+        lift = 1 + (ceplifter / 2.) * torch.sin(math.pi * torch.arange(numcep, device=device) / ceplifter)
         return lift * cepstra
     else:
         return cepstra
@@ -270,62 +275,64 @@ def stereo2mono(signal):
         return signal
 
 
-def calculate_frequencies(frame, samplerate):
+def calculate_frequencies(frames, samplerate):
     """
-    Calculate half frequencies (0, sample rate) within a frame.
-    :param frame: (framelen,)
+    Calculate half frequencies of the sample rate.
+    :param frame: (nframes, framelen)
     :param samplerate:
     :return: (framelen//2,) or or (framelen//2 - 1,)
     """
     # 半频
-    n = len(frame)
-    t = frame.shape[0] * 1.0 / samplerate
+    winlen = frames.shape[1]
+    t = winlen * 1.0 / samplerate
     if samplerate % 2 == 0:
-        return torch.arange(1, n // 2 + 1) / t
+        return torch.arange(1, winlen // 2 + 1, device=device) / t
     else:
-        return torch.arange(1, (n - 1) // 2 + 1) / t
+        return torch.arange(1, (winlen - 1) // 2 + 1, device=device) / t
 
 
-def calculate_energy(frame):
+def calculate_energy(frames):
     """
-    Calculate energy of a frame by rfft.
-    :param frame: (framelen,)
-    :return: (framelen//2,) or (framelen//2 - 1,) that equals to half frequencies
+    Calculate energy of each frame by rfft.
+    :param frame: (nframes, framelen)
+    :return: (nframes, framelen//2) or (nframes, framelen//2 - 1)
+             that equals to half frequencies
     """
-    mag = torch.norm(torch.rfft(frame, 1), dim=1)[1:]
-    energy = mag.T ** 2
+    mag = torch.norm(torch.rfft(frames, 1), dim=2)[:,1:]
+    energy = mag ** 2
     return energy
 
 
-def freq_energy(frame, samplerate):
+def freq_energy(frames, samplerate):
     """
-    Calculate a pair of (frequencies, energy) of a frame.
-    :param frame: (framelen,)
+    Calculate a pair of (frequencies, energy) of each frame.
+    :param frame: (nframes, framelen)
     :param samplerate:
-    :return: both (framelen//2,) or (framelen//2-1,)
+    :return: freq (framelen//2,) or (framelen//2-1,)
+            energy (nframes, framelen//2) or (nframes, framelen//2 - 1) corresponding to freq
     """
-    freq = calculate_frequencies(frame, samplerate)
-    energy = calculate_energy(frame)
+    freq = calculate_frequencies(frames, samplerate)
+    energy = calculate_energy(frames)
     return freq, energy
 
 
-def energy_ratio(frame, samplerate, lowfreq, highfreq):
+def energy_ratio(freq, energy, thresEnergy, lowfreq, highfreq):
     """
-    Calculate the ratio between energy of speech band and total energy for window.
-    :param frame: (framelen,)
-    :param samplerate:
+    Calculate the ratio between energy of speech band and total energy for a frame.
+    :param freq: (winlen//2)
+    :param energy: (nframes, winlen//2)
     :param lowfreq:
     :param highfreq:
-    :return: float
+    :return: (nframes,)
     """
-    freq, energy = freq_energy(frame, samplerate)
-    voice_energy = torch.sum(energy[(freq > lowfreq) & (freq < highfreq)])
-    full_energy = torch.sum(energy)
-    if full_energy == 0: full_energy = 2.220446049250313e-16  # 极小正数替换0
-    return voice_energy / full_energy
+    voice_energy = torch.sum(energy[:, (freq > lowfreq) & (freq < highfreq)], dim=1)
+    full_energy = torch.sum(energy, dim=1)
+    full_energy[full_energy == 0] = 2.220446049250313e-16  # 极小正数替换 0
+    detection = (torch.div(voice_energy, full_energy) >= thresEnergy).type(torch.float32)
+    return detection
 
 
-def smooth_detection(detect, winlen, speechlen):
+def smooth_detection(detection, winlen, speechlen):
     """
     Apply median filter with length of {speechlen} to smooth detected speech regions
     :param detect: (nframes,)
@@ -337,26 +344,26 @@ def smooth_detection(detect, winlen, speechlen):
     if medianwin % 2 == 0:
         medianwin -= 1
     mid = (medianwin - 1) // 2
-    y = torch.zeros((len(detect), medianwin), dtype=detect.dtype)
-    y[:, mid] = detect
+    y = torch.zeros((len(detection), medianwin), dtype=detection.dtype, device=device)
+    y[:, mid] = detection
     for i in range(mid):
         j = mid - i
-        y[j:, i] = detect[:-j]
-        y[:j, i] = detect[0]
-        y[:-j, -(i + 1)] = detect[j:]
-        y[-j:, -(i + 1)] = detect[-1]
-    medianEnergy = torch.median(y.type(torch.float), dim=1).values.type(torch.bool)
+        y[j:, i] = detection[:-j]
+        y[:j, i] = detection[0]
+        y[:-j, -(i + 1)] = detection[j:]
+        y[-j:, -(i + 1)] = detection[-1]
+    medianEnergy, _ = torch.median(y.type(torch.float), dim=1)
     return medianEnergy
 
 
-def is_speech(signal, samplerate=16000, winlen=0.02, hoplen=0.01, thresEnergy=0.6, speechlen=0.5,
-              lowfreq=300, highfreq=3000, preemph=0.97):
+def is_speech(wav, samplerate=16000, winlen=0.02, hoplen=0.01, thresEnergy=0.6, speechlen=0.5,
+              lowfreq=300, highfreq=3000, preemph=0.97, winfunc=lambda x: torch.ones((x,), device=device)):
     """
     Use signal energy to detect voice activity in PyTorch's Tensor.
     Detects speech regions based on ratio between speech band energy and total energy.
     Outputs are two tensor with the number of frames where the first output is start frame
         and the second output is to indicate voice activity.
-    :param signal: (time,)
+    :param wav: (time,)
     :param samplerate:
     :param winlen:
     :param hoplen:
@@ -365,18 +372,14 @@ def is_speech(signal, samplerate=16000, winlen=0.02, hoplen=0.01, thresEnergy=0.
     :param lowfreq:
     :param highfreq:
     :param preemph:
-    :return: (nframes,), (nframes)
+    :return: (nframes,), (nframes,), (time)
     """
-    if len(signal) < round(winlen * samplerate):
+    if len(wav) < round(winlen * samplerate):
         return torch.tensor([0]), torch.tensor([0])
-    signal = preemphasis(signal, preemph)
-    frames = framesig(signal, winlen * samplerate, hoplen * samplerate)
-    starts = torch.arange(0, frames.shape[0]) * round(hoplen * samplerate)
-    detection = torch.zeros(frames.shape[0], dtype=torch.bool)
-    for i, frame in enumerate(frames):
-        ratioEnergy = energy_ratio(frame, samplerate, lowfreq, highfreq)
-        detection[i] = ratioEnergy > thresEnergy
-    if speechlen:
-        detection = smooth_detection(detection, winlen, speechlen)
-    return starts, detection
-
+    wav = preemphasis(wav, preemph=preemph)
+    frames = framesig(wav, winlen*samplerate, hoplen*samplerate, winfunc)
+    freq, energy = freq_energy(frames, samplerate)
+    detection = energy_ratio(freq, energy, thresEnergy, lowfreq, highfreq)
+    detection = smooth_detection(detection, winlen, speechlen)
+    starts = torch.arange(0, frames.shape[0], device=device) * round(hoplen * samplerate)
+    return detection, starts, wav
